/**
 * Manages loading and playing of audio elements.
 * Its init() function should be called immediately upon successful project load.
 */
function AudioManager() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;

    // Object of all indices which are currently looping.
    var loopingIndices = {};

    // Object of all indices which may be currently playing.
    var playingIndices = {};

    // All currently in use indices referenced by URL.
    var urlToIndex = {};

    // Context used to play loaded audio.
    var context;

    // A gain which will always be the last chain in the audio link to which all other sounds connect.
    var masterVolumeGain;

    // A gain which is tuned to the effects volume level, and linked to the master volume gain.
    var effectsVolumeGain;

    // A gain which is tuned to the background music volume level, and linked to the master volume gain.
    var backgroundVolumeGain;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Sets the passed gain to the given volume.
     * @param {GainNode} gain The gain to set the volume of.
     * @param {int} volume The volume, ranging from 0 to 100.
     */
    function setVolume(gain, volume) {
        volume = volume / 100;
        gain.gain.setValueAtTime(volume, context.currentTime);
    }

    /**
     * Sets the master volume to the given amount.
     * @param {int} volume The master volume, ranging from 0 to 100.
     */
    function setMasterVolume(volume) {
        setVolume(masterVolumeGain, volume);
    }

    /**
     * Sets the effects volume to the given amount.
     * @param {int} volume The effects volume, ranging from 0 to 100.
     */
    function setEffectsVolume(volume) {
        setVolume(effectsVolumeGain, volume);
    }

    /**
     * Sets the background volume to the given amount.
     * @param {int} volume The background volume, ranging from 0 to 100.
     */
    function setBackgroundVolume(volume) {
        setVolume(backgroundVolumeGain, volume);
    }

    /**
     * Adds a workaround to a shitty google autoplay policy decision.
     * See https://developers.google.com/web/updates/2017/09/autoplay-policy-changes#webaudio
     */
    function googleWorkaround() {
        if (!context || !that.audioEnabled) {
            return;
        }

        var resumeAudio = function () {
            if (context.state === "suspended") {
                context.resume();
            }
            document.removeEventListener("click", resumeAudio);
        };

        document.addEventListener("click", resumeAudio);
    }

    /**
     * Creates the audio context, if audio is enabled.
     * Disables audio and logs a warning to the console if the web audio api is not supported.
     * May cause a warning message which should be ignored. (Thanks, Google. Fuck you.)
     */
    function createContext() {
        if (!that.audioEnabled) {
            return;
        }

        try {
            window.AudioContext = window.AudioContext || window.webkitAudioContext;
            context = new AudioContext();
        } catch (e) {
            console.log('Web Audio API is not supported in this browser');
            that.audioEnabled = false;
        }

        googleWorkaround();
    }

    /**
     * Creates the master volume gain, and connects it to the output.
     * Does nothing if there is no context or if audio is disabled.
     */
    function createMasterVolumeGain() {
        if (!context || !that.audioEnabled) {
            return;
        }
        masterVolumeGain = context.createGain();
        setMasterVolume(global.settings.getMasterVolume());
        masterVolumeGain.connect(context.destination);
    }

    /**
     * Creates the effects volume gain, and connects it to the master volume gain.
     * Does nothing if there is no context or if audio is disabled.
     */
    function createEffectsVolumeGain() {
        if (!context || !that.audioEnabled) {
            return;
        }
        effectsVolumeGain = context.createGain();
        setEffectsVolume(global.settings.getEffectsVolume());
        effectsVolumeGain.connect(masterVolumeGain);
    }

    /**
     * Creates the background music volume gain, and connects it to the output.
     * Does nothing if there is no context or if audio is disabled.
     */
    function createBackgroundVolumeGain() {
        if (!context || !that.audioEnabled) {
            return;
        }
        backgroundVolumeGain = context.createGain();
        setBackgroundVolume(global.settings.getBackgroundVolume());
        backgroundVolumeGain.connect(masterVolumeGain);
    }

    /**
     * Creats and loads an AudioFile with the given URL.
     * @param {string} url URL of the audio file to load.
     * @param {int} defaultVolume The default volume for the audio file, from 1 to 100.
     * @returns {AudioFile} The AudioFile, after it has started (but not necessarily completed) loading.
     */
    function loadAudioFile(url, defaultVolume) {
        var file = new AudioFile(context, url, masterVolumeGain, effectsVolumeGain, backgroundVolumeGain, defaultVolume);
        file.load();
        return file;
    }

    /**
     * Adds an AudioFile to the audioFiles array, adds a mapping from the url to the index in the urlToIndex object, and then returns the index.
     * @param {AudioFile} file The AudioFile to add.
     * @returns {int} The index of the AudioFile in the audioFiles array.
     */
    function addAudioFile(file) {
        var index = that.audioFiles.length;
        that.audioFiles.push(file);
        urlToIndex[file.url] = index;
        return index;
    }

    // PUBLIC MEMBERS \\        this.public = "test";

    // Array of all loaded (or load-in-progress) audio files.
    this.audioFiles = [];

    // Whether or not audio is enabled at all.
    this.audioEnabled = false;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Initializes the audio manager.
     */
    this.init = function () {
        that.audioEnabled = global.settings.getAudioEnabled();
        createContext();
        createMasterVolumeGain();
        createEffectsVolumeGain();
        createBackgroundVolumeGain();
        global.settings.masterVolumeListener.register(setMasterVolume);
        global.settings.effectsVolumeListener.register(setEffectsVolume);
        global.settings.backgroundVolumeListener.register(setBackgroundVolume);
    };

    /**
     * Asynchronously loads the audio file referenced by the given URL.
     * @param {string} url URL of the audio file to load.
     * @param {int} defaultVolume The default volume for the audio file, from 1 to 100.
     * @returns {int} An index to the sound for use in all further audio operations, or -1 if audio playing is disabled.
     *                This index may also be used to directly access the AudioFile, although all looping operations MUST happen via the AudioManager.
     */
    this.load = function (url, defaultVolume) {
        if (!that.audioEnabled) {
            return -1;
        }
        if (urlToIndex.hasOwnProperty(url)) {
            return urlToIndex[url];
        }
        return addAudioFile(loadAudioFile(url, defaultVolume));
    };

    /**
     * Plays the audio file referenced by the index a single time.
     * Does nothing if the index is out of bounds (such as if audio is disabled entirely).
     * Does nothing if the audio has not yet loaded.
     * @param {int} index A sound-index obtained by the load() function.
     * @returns {int} An index which can be used to refer to the audio track later. -1 if audio is disabled.
     */
    this.play = function (index) {
        if (index < 0 || index >= that.audioFiles.length) {
            return -1;
        }
        playingIndices[index] = true;
        return that.audioFiles[index].play();
    };

    /**
     * Plays the audio file referenced by the index a single time after it is loaded.
     * Plays instantly if the audio file has already loaded.
     * Does nothing if the index is out of bounds (such as if audio is disabled entirely).
     * @param {int} index A sound-index obtained by the load() function.
     * @returns {int} An index which can be used to refer to the audio track later. -1 if audio is disabled.
     */
    this.playWhenLoaded = function (index) {
        if (index < 0 || index >= that.audioFiles.length) {
            return -1;
        }
        playingIndices[index] = true;
        return that.audioFiles[index].playWhenLoaded();
    };

    /**
     * Stops a single playback of the given index.
     * @param {int} index Index of the audio file, returned on creation.
     * @param {int} track Index of the playback, returned on playing.
     */
    this.stop = function (index, track) {
        if (index < 0 || index >= that.audioFiles.length) {
            return;
        }
        that.audioFiles[index].stop(track);
    };

    /**
     * Fades a single playback of the given index.
     * @param {int} index Index of the audio file, returned on creation.
     * @param {int} track Index of the playback, returned on playing.
     */
    this.fade = function (index, track) {
        if (index < 0 || index >= that.audioFiles.length) {
            return;
        }
        that.audioFiles[index].fade(track);
    };

    /**
     * Cancels playing the audio file referenced by the index when loaded.
     * Can only be cancelled before the file has finished loaded.
     * If the file has already loaded, this function has no effect.
     * Does nothing if the index is out of bounds (such as if audio is disabled entirely).
     * @param {int} index A sound-index obtained by the load() function.
     */
    this.cancelPlayWhenLoaded = function (index) {
        if (index < 0 || index >= that.audioFiles.length) {
            return;
        }
        that.audioFiles[index].cancelPlayWhenLoaded();
    };

    /**
     * Plays the audio file referenced by the index over and over again.
     * Does nothing if the index is out of bounds (such as if audio is disabled entirely).
     * Does nothing if the audio file is already looping.
     * Waits for the audio to load before attempting to play it.
     * @param {int} index A sound-index obtained by the load() function.
     */
    this.loop = function (index) {
        if (index < 0 || index >= that.audioFiles.length) {
            return;
        }
        loopingIndices[index] = true;
        that.audioFiles[index].loop();
    };

    /**
     * Immediately stops the looping audio file referenced by the index.
     * Does nothing if the index is out of bounds (such as if audio is disabled entirely).
     * Does nothing if the audio file is not looping.
     * Has no effect on audio started by the play() function.
     * @param {int} index A sound-index obtained by the load() function.
     */
    this.stopLoop = function (index) {
        if (index < 0 || index >= that.audioFiles.length) {
            return;
        }
        delete loopingIndices[index];
        that.audioFiles[index].stopLoop();
    };

    /**
     * Smoothly fades out the looping audio file referenced by the index.
     * Does nothing if the index is out of bounds (such as if audio is disabled entirely).
     * Does nothing if the audio file is not looping.
     * Has no effect on audio started by the play() function.
     * @param {int} index A sound-index obtained by the load() function.
     */
    this.fadeLoop = function (index) {
        if (index < 0 || index >= that.audioFiles.length) {
            return;
        }
        delete loopingIndices[index];
        that.audioFiles[index].fadeLoop();
    };

    /**
     * Fades out all looping audio currently playing.
     */
    this.fadeAllLoopingAudio = function () {
        for (var index in loopingIndices) {
            if (loopingIndices.hasOwnProperty(index)) {
                that.fadeLoop(index);
            }
        }
        loopingIndices = {};
    };

    /**
     * Fades out all audio currently playing, looping or otherwise.
     */
    this.fadeAllAudio = function () {
        that.fadeAllLoopingAudio();
        for (var index in playingIndices) {
            if (playingIndices.hasOwnProperty(index)) {
                that.audioFiles[index].fadeAll();
            }
        }
        playingIndices = {};
    };
}

