/**
 * The Hotloader is capable of loading *all* scenes at once,
 * favoring a single, optimized loading period over several smaller ones at scene-transition time.
 * Note that the Titlescreen scene is not loaded by this class.
 */
function HotLoader() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;

    // Maps scene id to scene folder.
    var folderMap = {
        "Introduction": "",
        "Forest": "",
        "Home": "",
        "SekiWorkshop": "HamletDay2",
        "BreedingFarm": "HamletDay2",
        "FarmerHouse": "HamletDay2",
        "HamletWell" : "HamletDay2",
        "UndergroundCave" : "HamletDay2",
        "NightTwo" : "Varanar",
        "MorningThree" : "Varanar/DayThree",
        "CaravanConfinement" : "Enslavement",
        "RiversideMarsh" : "Deepwoods"
    };

    // Resolve function for loading promise.
    var loadResolve;
    // Reject function for loading promise.
    var loadReject;

    // ID of the div containing the hot load in progress warning.
    var hotLoadWarningId = "hotloadwarning";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Stores the resolve and reject functions for use later.
     * @param {function} resolve Load resolve function.
     * @param {function} reject Load reject function.
     */
    function storeLoadResolve(resolve, reject) {
        loadResolve = function () {
            $(`#${hotLoadWarningId}`).hide(400);
            resolve();
        };
        loadReject = function (reason) {
            $(`#${hotLoadWarningId}`).hide(400);
            reject(reason);
        };
    }

    /**
     * Constructs the path to the base folder of the scene with the given id.
     * @param {string} id The id of the scene for which to construct the folder path.
     * @returns {string} The base folder path for the scene with the given id.
     */
    function constructFolderPath(id) {
        var folder = folderMap[id] || "";
        if (folder) {
            return `scenes/${folder}/${id}/`;
        } else {
            return `scenes/${id}/`;
        }
    }

    /**
     * Constructs the path to the JSON file of the scene with the given id.
     * @param {string} id The id of the scene for which to construct the JSON file path.
     * @returns {string} The JSON file path for the scene with the given id.
     */
    function constructJsonPath(id) {
        return constructFolderPath(id) + `${id}.json`;
    }

    /**
     * For each entry in the folder map, constructs a path to the JSON file and adds it to an array.
     * @returns {array} The JSON path array.
     */
    function constructJsonPathArray() {
        var jsonPathArray = {};
        for (var id in folderMap) {
            jsonPathArray[id] = constructJsonPath(id);
        }
        return jsonPathArray;
    }

    /**
     * Transforms the results of fetching the JSON files into a single array containing nothing but the JSONs themselves.
     * @param {array of arrays} jsonResults The results of the JSON fetching operation.
     * @returns {array} An array containing the JSONs themselves as objects.
     */
    function jsonResultsToJsonArray(jsonResults) {
        var jsons = [];
        for (var index in jsonResults) {
            jsons.push(jsonResults[index][0]);
        }
        return jsons;
    }

    /**
     * Indexes JSON-objects using the ids of the scenes as keys.
     * @param {array} jsonPathArray The JSON paths array, which contains the keys to use.
     * @param {array} jsons The loaded json objects array, containing the json objects to index.
     * @returns {array} An indexed array with scene ids as keys, and JSON-objects as values.
     */
    function indexJsonArray(jsonPathArray, jsons) {
        var indexedJsons = {};
        var keys = Object.keys(jsonPathArray);
        for (var i = 0; i < keys.length && i < jsons.length; i++) {
            indexedJsons[keys[i]] = jsons[i];
        }
        return indexedJsons;
    }

    /**
     * Loads the javascript file with the given path by adding it to the DOM.
     * @param {string} path The path of the file to load (including extension).
     * @returns {Promise} A Promise which will fulfill when the file has been successfully loaded.
     */
    function loadJavaScript(path) {
        return new Promise(function (resolve, reject) {
            var script = document.createElement("script");
            script.type = `text/javascript`;
            script.onerror = reject;
            script.src = path;
            script.onload = resolve;
            document.body.appendChild(script); // Don't use jQuery for this.
        });
    }

    /**
     * Loads all of the JavaScript files required for the passed scene.
     * @param {string} id The id of the scene to load.
     * @param {array} json The JSON-loaded file-listing array.
     * @returns {Promise} A Promise which will fulfill when all files have been successfully loaded.
     */
    function loadSceneJavaScript(id, json) {
        var requestArray = [];
        var folder = constructFolderPath(id);
        for (var i = 0; i < json.length; i++) {
            var path = `${folder}${json[i]}.js`;
            requestArray.push(loadJavaScript(path));
        }
        return $.when.apply($, requestArray);
    }

    /**
     * Commences hotloading of all JavaScript files referenced by the indexedJsons array.
     * Calls loadResolve when done, or loadReject when something fails.
     * @param {array} indexedJsons The id-indexed array of objects containing all files to load.
     */
    function hotLoadJavaScript(indexedJsons) {
        var requestArray = [];
        for (var index in indexedJsons) {
            requestArray.push(loadSceneJavaScript(index, indexedJsons[index]));
        }
        $.when.apply($, requestArray).then(function () {
            loadResolve();
        }, function () {
            loadReject(`Failed to load scenes. Perhaps there's something wrong with your internet connection? Clear your browser cache and try again.`);
        });
    }

    /**
     * Starts hotloading by getting all JSON files of all scenes.
     * Proceeds to hotLoadJavaScript after all JSON files have been fetched.
     * Calls loadReject when JSON loading fails.
     * @param {array} jsonPathArray The array of paths to JSON files.
     */
    function hotLoadStart(jsonPathArray) {
        var requestArray = [];
        for (var id in jsonPathArray) {
            requestArray.push($.getJSON(jsonPathArray[id]));
        }
        $.when.apply($, requestArray).then(function () {
            var jsons = jsonResultsToJsonArray(arguments);
            var indexedJsons = indexJsonArray(jsonPathArray, jsons);
            hotLoadJavaScript(indexedJsons);
        }, function () {
            loadReject(`Failed to load scene-JSONs. Perhaps there's something wrong with your internet connection? Clear your browser cache and try again.`);
        });
    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.loaded = new Promise(storeLoadResolve);

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Hotloads all the required game files asynchronously by adding references to them to the DOM.
     */
    this.hotLoad = function () {
        $(`#${hotLoadWarningId}`).show(400);
        var jsonPathArray = constructJsonPathArray();
        hotLoadStart(jsonPathArray);
    };

    /**
     * Performs regular loading.
     */
    this.regularLoad = function () {
        loadResolve();
    };

    /**
     * Performs regular or hotloading depending on the settings.
     */
    this.hotLoadIfRequired = function () {
        if (global.settings.getHotLoad()) {
            that.hotLoad();
        } else {
            that.regularLoad();
        }
    };
}

